from __future__ import annotations

from typing import List

from fal.sdk import RunnerInfo

from ._utils import get_client
from .parser import FalClientParser


def runners_table(runners: List[RunnerInfo]):
    from rich.table import Table

    table = Table()
    table.add_column("Alias")
    table.add_column("Runner ID")
    table.add_column("In Flight Requests")
    table.add_column("Missing Leases")
    table.add_column("Expires In")
    table.add_column("Uptime")
    table.add_column("Revision")

    for runner in runners:
        external_metadata = runner.external_metadata
        present = external_metadata.get("present_in_group", True)

        num_leases_with_request = len(
            [
                lease
                for lease in external_metadata.get("leases", [])
                if lease.get("request_id") is not None
            ]
        )

        table.add_row(
            runner.alias,
            # Mark lost runners in red
            runner.runner_id if present else f"[red]{runner.runner_id}[/]",
            str(runner.in_flight_requests),
            str(runner.in_flight_requests - num_leases_with_request),
            (
                "N/A (active)"
                if runner.expiration_countdown is None
                else f"{runner.expiration_countdown}s"
            ),
            f"{runner.uptime} ({runner.uptime.total_seconds()}s)",
            runner.revision,
        )

    return table


def runners_requests_table(runners: list[RunnerInfo]):
    from rich.table import Table

    table = Table()
    table.add_column("Runner ID")
    table.add_column("Request ID")
    table.add_column("Caller ID")

    for runner in runners:
        for lease in runner.external_metadata.get("leases", []):
            if not (req_id := lease.get("request_id")):
                continue

            table.add_row(
                runner.runner_id,
                req_id,
                lease.get("caller_user_id") or "",
            )

    return table


def _kill(args):
    client = get_client(args.host, args.team)
    with client.connect() as connection:
        connection.kill_runner(args.id)


def _list(args):
    client = get_client(args.host, args.team)
    with client.connect() as connection:
        runners = connection.list_runners()
        args.console.print(f"Runners: {len(runners)}")
        args.console.print(runners_table(runners))

        requests_table = runners_requests_table(runners)
        args.console.print(f"Requests: {len(requests_table.rows)}")
        args.console.print(requests_table)


def _add_kill_parser(subparsers, parents):
    kill_help = "Kill a runner."
    parser = subparsers.add_parser(
        "kill",
        description=kill_help,
        help=kill_help,
        parents=parents,
    )
    parser.add_argument(
        "id",
        help="Runner ID.",
    )
    parser.set_defaults(func=_kill)


def _add_list_parser(subparsers, parents):
    list_help = "List runners."
    parser = subparsers.add_parser(
        "list",
        description=list_help,
        help=list_help,
        parents=parents,
    )
    parser.set_defaults(func=_list)


def add_parser(main_subparsers, parents):
    runners_help = "Manage fal runners."
    parser = main_subparsers.add_parser(
        "runners",
        description=runners_help,
        help=runners_help,
        parents=parents,
        aliases=["machine"],  # backwards compatibility
    )

    subparsers = parser.add_subparsers(
        title="Commands",
        metavar="command",
        required=True,
        parser_class=FalClientParser,
    )

    _add_kill_parser(subparsers, parents)
    _add_list_parser(subparsers, parents)



from http import HTTPStatus
from typing import Any, Dict, Optional, Union

import httpx

from ... import errors
from ...client import Client
from ...models.http_validation_error import HTTPValidationError
from ...models.typed_workflow import TypedWorkflow
from ...models.workflow_detail import WorkflowDetail
from ...types import Response


def _get_kwargs(
    *,
    client: Client,
    json_body: TypedWorkflow,
) -> Dict[str, Any]:
    url = "{}/workflows/".format(client.base_url)

    headers: Dict[str, str] = client.get_headers()
    cookies: Dict[str, Any] = client.get_cookies()

    json_json_body = json_body.to_dict()

    return {
        "method": "post",
        "url": url,
        "headers": headers,
        "cookies": cookies,
        "timeout": client.get_timeout(),
        "follow_redirects": client.follow_redirects,
        "json": json_json_body,
    }


def _parse_response(
    *, client: Client, response: httpx.Response
) -> Optional[Union[HTTPValidationError, WorkflowDetail]]:
    if response.status_code == HTTPStatus.CREATED:
        response_201 = WorkflowDetail.from_dict(response.json())

        return response_201
    if response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
        response_422 = HTTPValidationError.from_dict(response.json())

        return response_422
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Client, response: httpx.Response
) -> Response[Union[HTTPValidationError, WorkflowDetail]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    *,
    client: Client,
    json_body: TypedWorkflow,
) -> Response[Union[HTTPValidationError, WorkflowDetail]]:
    """Create Workflow

    Args:
        json_body (TypedWorkflow):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[HTTPValidationError, WorkflowDetail]]
    """

    kwargs = _get_kwargs(
        client=client,
        json_body=json_body,
    )

    response = httpx.request(
        verify=client.verify_ssl,
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    *,
    client: Client,
    json_body: TypedWorkflow,
) -> Optional[Union[HTTPValidationError, WorkflowDetail]]:
    """Create Workflow

    Args:
        json_body (TypedWorkflow):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[HTTPValidationError, WorkflowDetail]
    """

    return sync_detailed(
        client=client,
        json_body=json_body,
    ).parsed


async def asyncio_detailed(
    *,
    client: Client,
    json_body: TypedWorkflow,
) -> Response[Union[HTTPValidationError, WorkflowDetail]]:
    """Create Workflow

    Args:
        json_body (TypedWorkflow):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[HTTPValidationError, WorkflowDetail]]
    """

    kwargs = _get_kwargs(
        client=client,
        json_body=json_body,
    )

    async with httpx.AsyncClient(verify=client.verify_ssl) as _client:
        response = await _client.request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    *,
    client: Client,
    json_body: TypedWorkflow,
) -> Optional[Union[HTTPValidationError, WorkflowDetail]]:
    """Create Workflow

    Args:
        json_body (TypedWorkflow):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[HTTPValidationError, WorkflowDetail]
    """

    return (
        await asyncio_detailed(
            client=client,
            json_body=json_body,
        )
    ).parsed



from __future__ import annotations

from structlog.typing import EventDict, WrappedLogger


class AddUserIdProcessor:
    def __init__(self):
        from fal.auth import UserAccess

        self.user_access = UserAccess()

    def __call__(
        self, logger: WrappedLogger, method_name: str, event_dict: EventDict
    ) -> EventDict:
        """The structlog processor that sends the logged user id on every log"""
        user_id: str | None = None
        try:
            user_id = self.user_access.info.get("sub")
        except Exception:
            # logs are fail-safe, so any exception is safe to ignore
            # this is expected to happen only when user is logged out
            # or there's no internet connection
            pass
        event_dict["usr.id"] = user_id
        return event_dict



import os
import posixpath
from functools import cached_property
from typing import TYPE_CHECKING

from fsspec import AbstractFileSystem

if TYPE_CHECKING:
    import httpx

USER_AGENT = "fal-sdk/1.14.0 (python)"


class FalFileSystem(AbstractFileSystem):
    @cached_property
    def _client(self) -> "httpx.Client":
        from httpx import Client

        from fal.flags import REST_URL
        from fal.sdk import get_default_credentials

        creds = get_default_credentials()
        return Client(
            base_url=REST_URL,
            headers={
                **creds.to_headers(),
                "User-Agent": USER_AGENT,
            },
        )

    def _request(self, method, path, **kwargs):
        from fal.exceptions import FalServerlessException

        response = self._client.request(method, path, **kwargs)
        if response.status_code != 200:
            try:
                detail = response.json()["detail"]
            except Exception:
                detail = response.text
            raise FalServerlessException(detail)
        return response

    def _ls(self, path):
        response = self._request("GET", f"/files/list/{path}")
        files = response.json()
        return sorted(
            (
                {
                    "name": entry["path"],
                    "size": entry["size"],
                    "type": "file" if entry["is_file"] else "directory",
                    "mtime": entry["updated_time"],
                }
                for entry in files
            ),
            key=lambda x: x["name"],
        )

    def ls(self, path, detail=True, **kwargs):
        abs_path = "/" + path.lstrip("/")
        if abs_path in self.dircache:
            entries = self.dircache[abs_path]
        elif abs_path in ["/", "", "."]:
            entries = [
                {
                    "name": "/data",
                    "size": 0,
                    "type": "directory",
                    "mtime": 0,
                }
            ]
        else:
            entries = self._ls(abs_path)
        self.dircache[abs_path] = entries

        if detail:
            return entries

        return [entry["name"] for entry in entries]

    def info(self, path, **kwargs):
        parent = posixpath.dirname(path)
        entries = self.ls(parent, detail=True)
        for entry in entries:
            if entry["name"] == path:
                return entry
        raise FileNotFoundError(f"File not found: {path}")

    def get_file(self, rpath, lpath, **kwargs):
        if self.isdir(rpath):
            os.makedirs(lpath, exist_ok=True)
            return

        with open(lpath, "wb") as fobj:
            response = self._request("GET", f"/files/file/{rpath}")
            fobj.write(response.content)

    def put_file(self, lpath, rpath, mode="overwrite", **kwargs):
        if os.path.isdir(lpath):
            return

        with open(lpath, "rb") as fobj:
            self._request(
                "POST",
                f"/files/file/local/{rpath}",
                files={"file_upload": (posixpath.basename(lpath), fobj, "text/plain")},
            )
        self.dircache.clear()

    def put_file_from_url(self, url, rpath, mode="overwrite", **kwargs):
        self._request(
            "POST",
            f"/files/file/url/{rpath}",
            json={"url": url},
        )
        self.dircache.clear()

    def rm(self, path, **kwargs):
        self._request(
            "DELETE",
            f"/files/file/{path}",
        )
        self.dircache.clear()



from fal.cli.keys import _create, _list, _revoke
from fal.cli.main import parse_args


def test_create():
    args = parse_args(
        [
            "keys",
            "create",
            "--scope",
            "API",
            "--desc",
            "My test key",
        ]
    )
    assert args.func == _create
    assert args.scope == "API"
    assert args.desc == "My test key"


def test_list():
    args = parse_args(["keys", "list"])
    assert args.func == _list


def test_revoke():
    args = parse_args(["keys", "revoke", "my-key"])
    assert args.func == _revoke
    assert args.key_id == "my-key"



from __future__ import annotations

from typing import TYPE_CHECKING

import fal.cli.runners as runners

from ._utils import get_client
from .parser import FalClientParser

if TYPE_CHECKING:
    from fal.sdk import AliasInfo, ApplicationInfo


def _apps_table(apps: list[AliasInfo]):
    from rich.table import Table

    table = Table()
    table.add_column("Name", no_wrap=True)
    table.add_column("Revision")
    table.add_column("Auth")
    table.add_column("Min Concurrency")
    table.add_column("Max Concurrency")
    table.add_column("Concurrency Buffer")
    table.add_column("Max Multiplexing")
    table.add_column("Keep Alive")
    table.add_column("Request Timeout")
    table.add_column("Startup Timeout")
    table.add_column("Machine Type")
    table.add_column("Runners")
    table.add_column("Regions")

    for app in apps:
        table.add_row(
            app.alias,
            app.revision,
            app.auth_mode,
            str(app.min_concurrency),
            str(app.max_concurrency),
            str(app.concurrency_buffer),
            str(app.max_multiplexing),
            str(app.keep_alive),
            str(app.request_timeout),
            str(app.startup_timeout),
            " ".join(app.machine_types),
            str(app.active_runners),
            " ".join(app.valid_regions),
        )

    return table


def _list(args):
    client = get_client(args.host, args.team)
    with client.connect() as connection:
        apps = connection.list_aliases()

        if args.filter:
            apps = [app for app in apps if args.filter in app.alias]

        if args.sort_by_runners:
            apps.sort(key=lambda x: x.active_runners)
        else:
            apps.sort(key=lambda x: x.alias)

        table = _apps_table(apps)

    args.console.print(table)


def _add_list_parser(subparsers, parents):
    list_help = "List applications."
    parser = subparsers.add_parser(
        "list",
        description=list_help,
        help=list_help,
        parents=parents,
    )
    parser.add_argument(
        "--sort-by-runners",
        action="store_true",
        help="Sort by number of runners ascending",
    )
    parser.add_argument(
        "--filter",
        type=str,
        help="Filter applications by alias contents",
    )
    parser.set_defaults(func=_list)


def _app_rev_table(revs: list[ApplicationInfo]):
    from rich.table import Table

    table = Table()
    table.add_column("Revision", no_wrap=True)
    table.add_column("Min Concurrency")
    table.add_column("Max Concurrency")
    table.add_column("Max Multiplexing")
    table.add_column("Keep Alive")
    table.add_column("Request Timeout")
    table.add_column("Startup Timeout")
    table.add_column("Machine Type")
    table.add_column("Runners")
    table.add_column("Regions")
    table.add_column("Created")

    for rev in revs:
        table.add_row(
            rev.application_id,
            str(rev.min_concurrency),
            str(rev.max_concurrency),
            str(rev.max_multiplexing),
            str(rev.keep_alive),
            str(rev.request_timeout),
            str(rev.startup_timeout),
            " ".join(rev.machine_types),
            str(rev.active_runners),
            " ".join(rev.valid_regions),
            str(rev.created_at),
        )

    return table


def _list_rev(args):
    client = get_client(args.host, args.team)
    with client.connect() as connection:
        revs = connection.list_applications(args.app_name)
        table = _app_rev_table(revs)

    args.console.print(table)


def _add_list_rev_parser(subparsers, parents):
    list_help = "List application revisions."
    parser = subparsers.add_parser(
        "list-rev",
        description=list_help,
        help=list_help,
        parents=parents,
    )
    parser.add_argument(
        "app_name",
        nargs="?",
        help="Application name.",
    )
    parser.set_defaults(func=_list_rev)


def _scale(args):
    client = get_client(args.host, args.team)
    with client.connect() as connection:
        if (
            args.keep_alive is None
            and args.max_multiplexing is None
            and args.max_concurrency is None
            and args.min_concurrency is None
            and args.concurrency_buffer is None
            and args.request_timeout is None
            and args.startup_timeout is None
            and args.machine_types is None
            and args.regions is None
        ):
            args.console.log("No parameters for update were provided, ignoring.")
            return

        alias_info = connection.update_application(
            application_name=args.app_name,
            keep_alive=args.keep_alive,
            max_multiplexing=args.max_multiplexing,
            max_concurrency=args.max_concurrency,
            min_concurrency=args.min_concurrency,
            concurrency_buffer=args.concurrency_buffer,
            request_timeout=args.request_timeout,
            startup_timeout=args.startup_timeout,
            machine_types=args.machine_types,
            valid_regions=args.regions,
        )
        table = _apps_table([alias_info])

    args.console.print(table)


def _add_scale_parser(subparsers, parents):
    scale_help = "Scale application."
    parser = subparsers.add_parser(
        "scale",
        description=scale_help,
        help=scale_help,
        parents=parents,
    )
    parser.add_argument(
        "app_name",
        help="Application name.",
    )
    parser.add_argument(
        "--keep-alive",
        type=int,
        help="Keep alive (seconds).",
    )
    parser.add_argument(
        "--max-multiplexing",
        type=int,
        help="Maximum multiplexing",
    )
    parser.add_argument(
        "--max-concurrency",
        type=int,
        help="Maximum concurrency.",
    )
    parser.add_argument(
        "--min-concurrency",
        type=int,
        help="Minimum concurrency",
    )
    parser.add_argument(
        "--concurrency-buffer",
        type=int,
        help="Concurrency buffer",
    )
    parser.add_argument(
        "--request-timeout",
        type=int,
        help="Request timeout (seconds).",
    )
    parser.add_argument(
        "--startup-timeout",
        type=int,
        help="Startup timeout (seconds).",
    )
    parser.add_argument(
        "--machine-types",
        type=str,
        nargs="+",
        dest="machine_types",
        help="Machine types (pass several items to set multiple).",
    )
    parser.add_argument(
        "--regions",
        nargs="+",
        help="Valid regions (pass several items to set multiple).",
    )
    parser.set_defaults(func=_scale)


def _set_rev(args):
    client = get_client(args.host, args.team)
    with client.connect() as connection:
        connection.create_alias(args.app_name, args.app_rev, args.auth)


def _add_set_rev_parser(subparsers, parents):
    from fal.sdk import ALIAS_AUTH_MODES

    set_help = "Set application to a particular revision."
    parser = subparsers.add_parser(
        "set-rev",
        description=set_help,
        help=set_help,
        parents=parents,
    )
    parser.add_argument(
        "app_name",
        help="Application name.",
    )
    parser.add_argument(
        "app_rev",
        help="Application revision.",
    )
    parser.add_argument(
        "--auth",
        choices=ALIAS_AUTH_MODES,
        default="private",
        help="Application authentication mode.",
    )
    parser.set_defaults(func=_set_rev)


def _runners(args):
    client = get_client(args.host, args.team)
    with client.connect() as connection:
        alias_runners = connection.list_alias_runners(alias=args.app_name)

    runners_table = runners.runners_table(alias_runners)
    args.console.print(f"Runners: {len(alias_runners)}")
    # Drop the alias column, which is the first column
    runners_table.columns.pop(0)
    args.console.print(runners_table)

    requests_table = runners.runners_requests_table(alias_runners)
    args.console.print(f"Requests: {len(requests_table.rows)}")
    args.console.print(requests_table)


def _add_runners_parser(subparsers, parents):
    runners_help = "List application runners."
    parser = subparsers.add_parser(
        "runners",
        description=runners_help,
        help=runners_help,
        parents=parents,
    )
    parser.add_argument(
        "app_name",
        help="Application name.",
    )
    parser.set_defaults(func=_runners)


def _delete(args):
    client = get_client(args.host, args.team)
    with client.connect() as connection:
        connection.delete_alias(args.app_name)


def _add_delete_parser(subparsers, parents):
    delete_help = "Delete application."
    parser = subparsers.add_parser(
        "delete",
        description=delete_help,
        help=delete_help,
        parents=parents,
    )
    parser.add_argument(
        "app_name",
        help="Application name.",
    )
    parser.set_defaults(func=_delete)


def _delete_rev(args):
    client = get_client(args.host, args.team)
    with client.connect() as connection:
        connection.delete_application(args.app_rev)


def _add_delete_rev_parser(subparsers, parents):
    delete_help = "Delete application revision."
    parser = subparsers.add_parser(
        "delete-rev",
        description=delete_help,
        help=delete_help,
        parents=parents,
    )
    parser.add_argument(
        "app_rev",
        help="Application revision.",
    )
    parser.set_defaults(func=_delete_rev)


def add_parser(main_subparsers, parents):
    apps_help = "Manage fal applications."
    parser = main_subparsers.add_parser(
        "apps",
        aliases=["app"],
        description=apps_help,
        help=apps_help,
        parents=parents,
    )

    subparsers = parser.add_subparsers(
        title="Commands",
        metavar="command",
        required=True,
    )

    parents = [*parents, FalClientParser(add_help=False)]

    _add_list_parser(subparsers, parents)
    _add_list_rev_parser(subparsers, parents)
    _add_set_rev_parser(subparsers, parents)
    _add_scale_parser(subparsers, parents)
    _add_runners_parser(subparsers, parents)
    _add_delete_parser(subparsers, parents)
    _add_delete_rev_parser(subparsers, parents)



from __future__ import annotations

from ._base import (
    AppException,  # noqa: F401
    FalServerlessException,  # noqa: F401
    FieldException,  # noqa: F401
    RequestCancelledException,  # noqa: F401
)
from ._cuda import CUDAOutOfMemoryException  # noqa: F401



from __future__ import annotations

import errno
import hashlib
import os
import shutil
import subprocess
import sys
from contextlib import suppress
from pathlib import Path, PurePath
from tempfile import NamedTemporaryFile, TemporaryDirectory
from urllib.parse import urlparse
from urllib.request import Request, urlopen

FAL_PERSISTENT_DIR = PurePath("/data")
FAL_REPOSITORY_DIR = FAL_PERSISTENT_DIR / ".fal" / "repos"
FAL_MODEL_WEIGHTS_DIR = FAL_PERSISTENT_DIR / ".fal" / "model_weights"


# TODO: how can we randomize the user agent to avoid being blocked?
TEMP_HEADERS = {
    "User-Agent": (
        "Mozilla/5.0 (Macintosh; "
        "Intel Mac OS X 10.8; rv:21.0) Gecko/20100101 Firefox/21.0"
    ),
}


class DownloadError(Exception):
    pass


def _hash_url(url: str) -> str:
    """Hashes a URL using SHA-256.

    Args:
        url: The URL to be hashed.

    Returns:
        A string representing the hashed URL.
    """
    return hashlib.sha256(url.encode("utf-8")).hexdigest()


def _get_remote_file_properties(
    url: str,
    request_headers: dict[str, str] | None = None,
) -> tuple[str, int]:
    """Retrieves the file name and content length of a remote file.

    This function sends an HTTP request to the remote URL and retrieves the
    "Content-Disposition" header and the "Content-Length" header from the response.
    The "Content-Disposition" header contains the file name of the remote file, while
    the "Content-Length" header contains the expected content length of the remote
    file.

    If the "Content-Disposition" header is not available, the function attempts to
    extract the file name from the URL's path component. If the URL does not contain a
    path component, the function generates a hashed value of the URL using SHA-256 and
    uses it as the file name.

    If the "Content-Length" header is not available, the function returns -1 as the
    content length.

    Args:
        url: The URL of the remote file.
        request_headers: A dictionary containing additional headers to be included in
            the HTTP request.

    Returns:
        A tuple containing the file name and the content length of the remote file.
    """
    headers = {**TEMP_HEADERS, **(request_headers or {})}
    request = Request(url, headers=headers)

    with urlopen(request) as response:
        file_name = response.headers.get_filename()
        content_length = int(response.headers.get("Content-Length", -1))

    if not file_name:
        parsed_url = urlparse(url)

        if parsed_url.scheme == "data":
            file_name = _hash_url(url)
        else:
            url_path = parsed_url.path
            file_name = Path(url_path).name or _hash_url(url)

    # file name can still contain a forward slash if the server returns a relative path
    file_name = Path(file_name).name

    return file_name, content_length


def _file_content_length_matches(
    url: str, file_path: Path, request_headers: dict[str, str] | None = None
) -> bool:
    """Check if the local file's content length matches the expected remote
    file's content length.

    This function compares the content length of a local file to the expected content
    length of a remote file, which is determined by sending an HTTP request to the
    remote URL. If the content lengths match, the function returns `True`, indicating
    that the local file's content matches the expected remote content. If the content
    lengths do not match or information about the content length is unavailable, the
    function returns `False`.

    Args:
        url: The URL of the remote file.
        file_path: The local path to the file being compared.
        request_headers: A dictionary containing additional headers to be included in
            the HTTP request.

    Returns:
        bool: `True` if the local file's content length matches the remote file's
            content length, `False` otherwise.
    """
    local_file_content_length = file_path.stat().st_size
    remote_file_content_length = _get_remote_file_properties(url, request_headers)[1]

    return local_file_content_length == remote_file_content_length


def download_file(
    url: str,
    target_dir: str | Path,
    *,
    force: bool = False,
    request_headers: dict[str, str] | None = None,
    filesize_limit: int | None = None,
) -> Path:
    """Downloads a file from the specified URL to the target directory.

    The function downloads the file from the given URL and saves it in the specified
    target directory, provided it is below the given filesize limit.

    It also checks whether the local file already exists and whether its content length
    matches the expected content length from the remote file. If the local file already
    exists and its content length matches the expected content length from the remote
    file, the existing file is returned without re-downloading it.

    If the file needs to be downloaded or if an existing file's content length does not
    match the expected length, the function proceeds to download and save the file. It
    ensures that the target directory exists and handles any errors that may occur
    during the download process, raising a `DownloadError` if necessary.

    Parameters:
        url: The URL of the file to be downloaded.
        target_dir: The directory where the downloaded file will be saved. If it's not
            an absolute path, it's treated as a relative directory to "/data".
        force: If `True`, the file is downloaded even if it already exists locally and
            its content length matches the expected content length from the remote file.
            Defaults to `False`.
        request_headers: A dictionary containing additional headers to be included in
            the HTTP request. Defaults to `None`.
        filesize_limit: An integer specifying the maximum downloadable size,
            in megabytes. Defaults to `None`.


    Returns:
        A Path object representing the full path to the downloaded file.

    Raises:
        ValueError: If the provided `file_name` contains a forward slash ('/').
        DownloadError: If an error occurs during the download process.
    """
    ONE_MB = 1024**2

    try:
        file_name, expected_filesize = _get_remote_file_properties(url, request_headers)
    except Exception as e:
        print(f"Got error: {e}")
        raise DownloadError(f"Failed to get remote file properties for {url}") from e

    expected_filesize_mb = expected_filesize / ONE_MB

    if filesize_limit is not None and expected_filesize_mb > filesize_limit:
        raise DownloadError(
            f"""File to be downloaded is of size {expected_filesize_mb},
                which is over the limit of {filesize_limit}"""
        )

    if "/" in file_name:
        raise ValueError(f"File name '{file_name}' cannot contain a slash.")

    target_dir_path = Path(target_dir)

    # If target_dir is not an absolute path, use "/data" as the relative directory
    if not target_dir_path.is_absolute():
        target_dir_path = Path(FAL_PERSISTENT_DIR / target_dir_path)  # type: ignore[assignment]

    target_path = target_dir_path.resolve() / file_name

    if (
        target_path.exists()
        and _file_content_length_matches(url, target_path, request_headers)
        and not force
    ):
        return target_path

    if force:
        print(f"File already exists. Forcing download of {url} to {target_path}")
    else:
        print(f"Downloading {url} to {target_path}")

    # Make sure the directory exists
    target_path.parent.mkdir(parents=True, exist_ok=True)

    try:
        _download_file_python(
            url=url,
            target_path=target_path,
            request_headers=request_headers,
            filesize_limit=filesize_limit,
        )
    except Exception as e:
        msg = f"Failed to download {url} to {target_path}"

        target_path.unlink(missing_ok=True)

        raise DownloadError(msg) from e

    return target_path


def _download_file_python(
    url: str,
    target_path: Path | str,
    request_headers: dict[str, str] | None = None,
    filesize_limit: int | None = None,
) -> Path:
    """Download a file from a given URL and save it to a specified path using a
    Python interface.

    Args:
        url: The URL of the file to be downloaded.
        target_path: The path where the downloaded file will be saved.
        request_headers: A dictionary containing additional headers to be included in
            the HTTP request. Defaults to `None`.
        filesize_limit: A integer value specifying how many megabytes can be
            downloaded at maximum. Defaults to `None`.

    Returns:
        The path where the downloaded file has been saved.
    """
    basename = os.path.basename(target_path)
    # NOTE: using the same directory to avoid potential copies across temp fs and target
    # fs, and also to be able to atomically rename a downloaded file into place.
    with NamedTemporaryFile(
        delete=False,
        dir=os.path.dirname(target_path),
        prefix=f"{basename}.tmp",
    ) as temp_file:
        try:
            file_path = temp_file.name

            for progress, total_size in _stream_url_data_to_file(
                url,
                temp_file.name,
                request_headers=request_headers,
                filesize_limit=filesize_limit,
            ):
                if total_size:
                    progress_msg = f"Downloading {url} ... {progress:.2%}"
                else:
                    progress_msg = f"Downloading {url} ... {progress:.2f} MB"

                print(progress_msg, end="\r\n")

            # NOTE: Atomically renaming the file into place when the file is downloaded
            # completely.
            #
            # Since the file used is temporary, in a case of an interruption, the
            # downloaded content will be lost. So, it is safe to redownload the file in
            # such cases.
            os.rename(file_path, target_path)

        finally:
            Path(temp_file.name).unlink(missing_ok=True)

    return Path(target_path)


def _stream_url_data_to_file(
    url: str,
    file_path: str,
    chunk_size_in_mb: int = 64,
    request_headers: dict[str, str] | None = None,
    filesize_limit: int | None = None,
):
    """Download data from a URL and stream it to a file.

    Note:
        - This function sets a User-Agent header to mimic a web browser to
            prevent issues with some websites.
        - It downloads the file in chunks to save memory and ensures the file
            is only moved when the download is complete.

    Args:
        request: The Request object representing the URL to download from.
        file_path: The path to the file where the downloaded data will be saved.
        chunk_size_in_mb: The size of each download chunk in megabytes.
            Defaults to 64.
        request_headers: A dictionary containing additional headers to be included in
            the HTTP request. Defaults to `None`.
        filesize_limit: An integer specifying how many megabytes can be
            downloaded at maximum. Defaults to `None`.

    Yields:
        A tuple containing two elements:
        - float: The progress of the download as a percentage (0.0 to 1.0) if
            the total size is known. Else, equals to the downloaded size in MB.
        - int: The total size of the downloaded content in bytes. If the total
            size is not known (e.g., the server doesn't provide a
            'content-length' header), the second element is 0.
    """
    ONE_MB = 1024**2

    headers = {**TEMP_HEADERS, **(request_headers or {})}
    request = Request(url, headers=headers)

    received_size = 0
    total_size = 0

    with urlopen(request) as response, open(file_path, "wb") as f_stream:
        total_size = int(response.headers.get("content-length", total_size))
        while data := response.read(chunk_size_in_mb * ONE_MB):
            f_stream.write(data)

            received_size = f_stream.tell()
            if filesize_limit is not None and received_size > filesize_limit:
                raise DownloadError(
                    f"""Attempted to download more data {received_size}
                        than the set limit of {filesize_limit}"""
                )

            if total_size:
                progress = received_size / total_size
            else:
                progress = received_size / ONE_MB
            yield progress, total_size

    # Check if received size matches the expected total size
    if total_size and received_size < total_size:
        raise DownloadError("Received less data than expected from the server.")


def download_model_weights(
    url: str, force: bool = False, request_headers: dict[str, str] | None = None
) -> Path:
    """Downloads model weights from the specified URL and saves them to a
    predefined directory.

    This function is specifically designed for downloading model weights and stores
    them in a predefined directory.

    It calls the `download_file` function with the provided
    URL and the target directory set to a pre-defined location for model weights.
    The downloaded model weights are saved in this directory, and the function returns
    the full path to the downloaded weights file.

    Args:
        url: The URL from which the model weights will be downloaded.
        force: If `True`, the model weights are downloaded even if they already exist
            locally and their content length matches the expected content length from
            the remote file. Defaults to `False`.
        request_headers: A dictionary containing additional headers to be included in
            the HTTP request. Defaults to `None`.

    Returns:
        A Path object representing the full path to the downloaded model weights.
    """
    # This is not a protected path, so the user may change stuff internally
    weights_dir = Path(FAL_MODEL_WEIGHTS_DIR / _hash_url(url))

    if weights_dir.exists() and not force:
        try:
            weights_path = next(weights_dir.glob("*"))
            return weights_path
        # The model weights directory is empty, so we need to download the weights
        except StopIteration:
            pass

    return download_file(
        url,
        target_dir=weights_dir,
        force=force,
        request_headers=request_headers,
    )


def clone_repository(
    https_url: str,
    *,
    commit_hash: str | None = None,
    target_dir: str | Path | None = None,
    repo_name: str | None = None,
    force: bool = False,
    include_to_path: bool = False,
) -> Path:
    """Clones a Git repository from the specified HTTPS URL into a local
    directory.

    This function clones a Git repository from the specified HTTPS URL into a local
    directory. It can also checkout a specific commit if the `commit_hash` is provided.

    If a custom `target_dir` or `repo_name` is not specified, a predefined directory is
    used for the target directory, and the repository name is determined from the URL.

    Args:
        https_url: The HTTPS URL of the Git repository to be cloned.
        commit_hash: The commit hash to checkout after cloning.
        target_dir: The directory where the repository will be saved.
            If not provided, a predefined directory is used.
        repo_name: The name to be used for the cloned repository directory.
            If not provided, the repository's name from the URL is used.
        force: If `True`, the repository is cloned even if it already exists locally
            and its commit hash matches the provided commit hash. Defaults to `False`.
        include_to_path: If `True`, the cloned repository is added to the `sys.path`.
            Defaults to `False`.

    Returns:
        A Path object representing the full path to the cloned Git repository.
    """
    target_dir = target_dir or FAL_REPOSITORY_DIR  # type: ignore[assignment]

    if repo_name is None:
        repo_name = Path(https_url).stem
        if commit_hash:
            repo_name += f"-{commit_hash[:8]}"

    local_repo_path = Path(target_dir) / repo_name  # type: ignore[arg-type]

    if local_repo_path.exists():
        local_repo_commit_hash = _get_git_revision_hash(local_repo_path)
        if local_repo_commit_hash == commit_hash and not force:
            if include_to_path:
                __add_local_path_to_sys_path(local_repo_path)
            return local_repo_path
        else:
            if local_repo_commit_hash != commit_hash:
                print(
                    f"Local repository '{local_repo_path}' has a different commit hash "
                    f"({local_repo_commit_hash}) than the one provided ({commit_hash})."
                )
            elif force:
                print(
                    f"Local repository '{local_repo_path}' already exists. "
                    f"Forcing re-download."
                )
            print(f"Removing the existing repository: {local_repo_path} ")
            with TemporaryDirectory(
                dir=target_dir, suffix=f"{local_repo_path.name}.tmp.old"
            ) as tmp_dir:
                with suppress(FileNotFoundError):
                    # repository might be already deleted by another worker
                    os.rename(local_repo_path, tmp_dir)
                    # sometimes seeing FileNotFoundError even here on juicefs
                    shutil.rmtree(tmp_dir)

    # NOTE: using the target_dir to be able to avoid potential copies across temp fs
    # and target fs, and also to be able to atomically rename repo_name dir into place
    # when we are done setting it up.
    os.makedirs(target_dir, exist_ok=True)  # type: ignore[arg-type]
    with TemporaryDirectory(
        dir=target_dir,
        suffix=f"{local_repo_path.name}.tmp",
    ) as temp_dir:
        try:
            print(f"Cloning the repository '{https_url}' .")

            # Clone with disabling the logs and advices for detached HEAD state.
            clone_command = [
                "git",
                "clone",
                "--recursive",
                https_url,
                temp_dir,
            ]
            subprocess.check_call(clone_command)

            if commit_hash:
                checkout_command = ["git", "checkout", commit_hash]
                subprocess.check_call(checkout_command, cwd=temp_dir)

            # NOTE: Atomically renaming the repository directory into place when the
            # clone and checkout are done.
            try:
                os.rename(temp_dir, local_repo_path)
            except OSError as error:
                shutil.rmtree(temp_dir)

                # someone beat us to it, assume it's good
                if error.errno != errno.ENOTEMPTY:
                    raise
                print(f"{local_repo_path} already exists, skipping rename")

        except Exception as error:
            print(f"{error}\nFailed to clone repository '{https_url}' .")
            raise error

    if include_to_path:
        __add_local_path_to_sys_path(local_repo_path)

    return local_repo_path


def __add_local_path_to_sys_path(local_path: Path | str):
    local_path_str = str(local_path)

    if local_path_str not in sys.path:
        sys.path.insert(0, local_path_str)


def _get_git_revision_hash(repo_path: Path) -> str:
    import subprocess

    try:
        return subprocess.check_output(
            ["git", "rev-parse", "HEAD"],
            cwd=repo_path,
            text=True,
            stderr=subprocess.STDOUT,
        ).strip()
    except subprocess.CalledProcessError as error:
        if "not a git repository" in error.output:
            print(f"Repository '{repo_path}' is not a git repository.")
            return ""

        print(
            f"{error}\nFailed to get the commit hash of the repository '{repo_path}' ."
        )
        raise error



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar

import attr

if TYPE_CHECKING:
    from ..models.workflow_contents import WorkflowContents


T = TypeVar("T", bound="TypedWorkflow")


@attr.s(auto_attribs=True)
class TypedWorkflow:
    """
    Attributes:
        name (str):
        title (str):
        contents (WorkflowContents):
        is_public (bool):
    """

    name: str
    title: str
    contents: "WorkflowContents"
    is_public: bool
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        name = self.name
        title = self.title
        contents = self.contents.to_dict()

        is_public = self.is_public

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
                "title": title,
                "contents": contents,
                "is_public": is_public,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.workflow_contents import WorkflowContents

        d = src_dict.copy()
        name = d.pop("name")

        title = d.pop("title")

        contents = WorkflowContents.from_dict(d.pop("contents"))

        is_public = d.pop("is_public")

        typed_workflow = cls(
            name=name,
            title=title,
            contents=contents,
            is_public=is_public,
        )

        typed_workflow.additional_properties = d
        return typed_workflow

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

import attr

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.workflow_contents import WorkflowContents


T = TypeVar("T", bound="TypedWorkflowUpdate")


@attr.s(auto_attribs=True)
class TypedWorkflowUpdate:
    """
    Attributes:
        name (Union[Unset, str]):
        title (Union[Unset, str]):
        contents (Union[Unset, WorkflowContents]):
        is_public (Union[Unset, bool]):
    """

    name: Union[Unset, str] = UNSET
    title: Union[Unset, str] = UNSET
    contents: Union[Unset, "WorkflowContents"] = UNSET
    is_public: Union[Unset, bool] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        name = self.name
        title = self.title
        contents: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.contents, Unset):
            contents = self.contents.to_dict()

        is_public = self.is_public

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if name is not UNSET:
            field_dict["name"] = name
        if title is not UNSET:
            field_dict["title"] = title
        if contents is not UNSET:
            field_dict["contents"] = contents
        if is_public is not UNSET:
            field_dict["is_public"] = is_public

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.workflow_contents import WorkflowContents

        d = src_dict.copy()
        name = d.pop("name", UNSET)

        title = d.pop("title", UNSET)

        _contents = d.pop("contents", UNSET)
        contents: Union[Unset, WorkflowContents]
        if isinstance(_contents, Unset):
            contents = UNSET
        else:
            contents = WorkflowContents.from_dict(_contents)

        is_public = d.pop("is_public", UNSET)

        typed_workflow_update = cls(
            name=name,
            title=title,
            contents=contents,
            is_public=is_public,
        )

        typed_workflow_update.additional_properties = d
        return typed_workflow_update

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from __future__ import annotations

from rich.table import Table

from fal.config import Config


def _list(args):
    table = Table()
    table.add_column("Default")
    table.add_column("Profile")
    table.add_column("Settings")

    config = Config()
    for profile in config.profiles():
        table.add_row(
            "*" if profile == config._profile else "",
            profile,
            ", ".join(key for key in config._config[profile]),
        )

    args.console.print(table)


def _set(args):
    with Config().edit() as config:
        config.set_internal("profile", args.PROFILE)
        args.console.print(f"Default profile set to [cyan]{args.PROFILE}[/].")
        config.profile = args.PROFILE
        if not config.get("key"):
            args.console.print(
                "No key set for profile. Use [bold]fal profile key[/] to set a key."
            )


def _unset(args, config: Config | None = None):
    config = config or Config()

    with config.edit() as config:
        config.profile = None
        args.console.print("Default profile unset.")


def _key_set(args):
    while True:
        key = input("Enter the key: ")
        if ":" in key:
            break
        args.console.print(
            "[red]Invalid key. The key must be in the format [bold]key:value[/].[/]"
        )

    with Config().edit() as config:
        config.set("key", key)
        args.console.print(f"Key set for profile [cyan]{config.profile}[/].")


def _host_set(args):
    with Config().edit() as config:
        config.set("host", args.HOST)
        args.console.print(f"Fal host set to [cyan]{args.HOST}[/].")


def _delete(args):
    with Config().edit() as config:
        if config.profile == args.PROFILE:
            config.set_internal("profile", None)

        config.delete_profile(args.PROFILE)
        args.console.print(f"Profile [cyan]{args.PROFILE}[/] deleted.")


def add_parser(main_subparsers, parents):
    auth_help = "Profile management."
    parser = main_subparsers.add_parser(
        "profile",
        aliases=["profiles"],
        description=auth_help,
        help=auth_help,
        parents=parents,
    )

    subparsers = parser.add_subparsers(
        title="Commands",
        metavar="command",
        dest="cmd",
        required=True,
    )

    list_help = "List all profiles."
    list_parser = subparsers.add_parser(
        "list",
        description=list_help,
        help=list_help,
        parents=parents,
    )
    list_parser.set_defaults(func=_list)

    set_help = "Set default profile."
    set_parser = subparsers.add_parser(
        "set",
        description=set_help,
        help=set_help,
        parents=parents,
    )
    set_parser.add_argument(
        "PROFILE",
        help="Profile name.",
    )
    set_parser.set_defaults(func=_set)

    unset_help = "Unset default profile."
    unset_parser = subparsers.add_parser(
        "unset",
        description=unset_help,
        help=unset_help,
        parents=parents,
    )
    unset_parser.set_defaults(func=_unset)

    key_set_help = "Set key for profile."
    key_set_parser = subparsers.add_parser(
        "key",
        description=key_set_help,
        help=key_set_help,
        parents=parents,
    )
    key_set_parser.set_defaults(func=_key_set)

    host_set_help = "Set fal host."
    host_set_parser = subparsers.add_parser(
        "host",
        description=host_set_help,
        help=host_set_help,
        parents=parents,
    )
    host_set_parser.add_argument(
        "HOST",
        help="Fal host.",
    )
    host_set_parser.set_defaults(func=_host_set)

    delete_help = "Delete profile."
    delete_parser = subparsers.add_parser(
        "delete",
        description=delete_help,
        help=delete_help,
        parents=parents,
    )
    delete_parser.add_argument(
        "PROFILE",
        help="Profile name.",
    )
    delete_parser.set_defaults(func=_delete)



from __future__ import annotations

import asyncio
import inspect
import json
import os
import queue
import re
import threading
import time
import typing
from contextlib import asynccontextmanager, contextmanager
from dataclasses import dataclass
from typing import Any, Callable, ClassVar, Literal, TypeVar

import fastapi
import grpc.aio as async_grpc
import httpx

from fal._serialization import include_modules_from
from fal.api import (
    SERVE_REQUIREMENTS,
    BaseServable,
    IsolatedFunction,
    RouteSignature,
)
from fal.api import (
    function as fal_function,
)
from fal.exceptions import FalServerlessException, RequestCancelledException
from fal.logging import get_logger
from fal.toolkit.file import request_lifecycle_preference
from fal.toolkit.file.providers.fal import LIFECYCLE_PREFERENCE

REALTIME_APP_REQUIREMENTS = ["websockets", "msgpack"]
REQUEST_ID_KEY = "x-fal-request-id"

EndpointT = TypeVar("EndpointT", bound=Callable[..., Any])
logger = get_logger(__name__)


async def _call_any_fn(fn, *args, **kwargs):
    if inspect.iscoroutinefunction(fn):
        return await fn(*args, **kwargs)
    else:
        return fn(*args, **kwargs)


async def open_isolate_channel(address: str) -> async_grpc.Channel | None:
    channel = async_grpc.insecure_channel(
        address,
        options=[
            ("grpc.max_send_message_length", -1),
            ("grpc.max_receive_message_length", -1),
            ("grpc.min_reconnect_backoff_ms", 0),
            ("grpc.max_reconnect_backoff_ms", 100),
            ("grpc.dns_min_time_between_resolutions_ms", 100),
        ],
    )

    try:
        channel_status = channel.channel_ready()

        await asyncio.wait_for(channel_status, timeout=1)
    except asyncio.TimeoutError:
        await channel.close(None)
        print("[DEBUG] Timed out trying to connect to local isolate")
        return None

    return channel


async def _set_logger_labels(
    logger_labels: dict[str, str], channel: async_grpc.Channel
):
    try:
        import sys

        from isolate.server import definitions

        # Flush any prints that were buffered before setting the logger labels
        sys.stderr.flush()
        sys.stdout.flush()

        isolate = definitions.IsolateStub(channel)
        isolate_request = definitions.SetMetadataRequest(
            # TODO: when submit is shipped, get task_id from an env var
            task_id="RUN",
            metadata=definitions.TaskMetadata(logger_labels=logger_labels),
        )
        res = isolate.SetMetadata(isolate_request)
        code = await res.code()
        assert str(code) == "StatusCode.OK", str(code)
    except BaseException:
        # NOTE hiding this for now to not print on every request
        # logger.debug("Failed to set logger labels", exc_info=True)
        pass


def wrap_app(cls: type[App], **kwargs) -> IsolatedFunction:
    include_modules_from(cls)

    def initialize_and_serve():
        app = cls()
        app.serve()

    metadata = {}
    app = cls(_allow_init=True)

    metadata["openapi"] = app.openapi()

    routes = app.collect_routes()
    realtime_app = any(route.is_websocket for route in routes)

    kind = cls.host_kwargs.pop("kind", "virtualenv")
    if kind == "container":
        cls.host_kwargs.pop("resolver", None)

    wrapper = fal_function(
        kind,
        requirements=cls.requirements,
        local_python_modules=cls.local_python_modules,
        machine_type=cls.machine_type,
        num_gpus=cls.num_gpus,
        **cls.host_kwargs,
        **kwargs,
        metadata=metadata,
        exposed_port=8080,
        serve=False,
    )
    fn = wrapper(initialize_and_serve)
    fn.options.add_requirements(SERVE_REQUIREMENTS)
    if realtime_app:
        fn.options.add_requirements(REALTIME_APP_REQUIREMENTS)

    return fn


@dataclass
class AppClientError(FalServerlessException):
    message: str
    status_code: int


class EndpointClient:
    def __init__(self, url, endpoint, signature, timeout: int | None = None):
        self.url = url
        self.endpoint = endpoint
        self.signature = signature
        self.timeout = timeout

        annotations = endpoint.__annotations__ or {}
        self.return_type = annotations.get("return") or None

    def __call__(self, data):
        with httpx.Client() as client:
            url = self.url + self.signature.path
            resp = client.post(
                self.url + self.signature.path,
                json=data.dict() if hasattr(data, "dict") else dict(data),
                timeout=self.timeout,
            )
            if not resp.is_success:
                # allow logs to be printed before raising the exception
                time.sleep(1)
                raise AppClientError(
                    f"Failed to POST {url}: {resp.status_code} {resp.text}",
                    status_code=resp.status_code,
                )
            resp_dict = resp.json()

        if not self.return_type:
            return resp_dict

        return self.return_type(**resp_dict)


class AppClient:
    def __init__(
        self,
        cls,
        url,
        timeout: int | None = None,
    ):
        self.url = url
        self.cls = cls

        for name, endpoint in inspect.getmembers(cls, inspect.isfunction):
            signature = getattr(endpoint, "route_signature", None)
            if signature is None:
                continue
            endpoint_client = EndpointClient(
                self.url,
                endpoint,
                signature,
                timeout=timeout,
            )
            setattr(self, name, endpoint_client)

    @classmethod
    @contextmanager
    def connect(cls, app_cls):
        app = wrap_app(app_cls)
        info = app.spawn()
        _shutdown_event = threading.Event()

        def _print_logs():
            while not _shutdown_event.is_set():
                try:
                    log = info.logs.get(timeout=0.1)
                except queue.Empty:
                    continue
                print(log)

        _log_printer = threading.Thread(target=_print_logs, daemon=True)
        _log_printer.start()

        try:
            with httpx.Client() as client:
                retries = 100
                for _ in range(retries):
                    url = info.url + "/health"
                    resp = client.get(url, timeout=60)

                    if resp.is_success:
                        break
                    elif resp.status_code not in (500, 404):
                        raise AppClientError(
                            f"Failed to GET {url}: {resp.status_code} {resp.text}",
                            status_code=resp.status_code,
                        )
                    time.sleep(0.1)

            client = cls(app_cls, info.url)
            yield client
        finally:
            info.stream.cancel()
            _shutdown_event.set()
            _log_printer.join()

    def health(self):
        with httpx.Client() as client:
            resp = client.get(self.url + "/health")
            resp.raise_for_status()
            return resp.json()


PART_FINDER_RE = re.compile(r"[A-Z][a-z]*")


def _to_fal_app_name(name: str) -> str:
    # Convert MyGoodApp into my-good-app
    return "-".join(part.lower() for part in PART_FINDER_RE.findall(name))


def _print_python_packages() -> None:
    from importlib.metadata import distributions

    packages = [f"{dist.metadata['Name']}=={dist.version}" for dist in distributions()]

    print("[debug] Python packages installed:", ", ".join(packages))


class App(BaseServable):
    requirements: ClassVar[list[str]] = []
    local_python_modules: ClassVar[list[str]] = []
    machine_type: ClassVar[str] = "S"
    num_gpus: ClassVar[int | None] = None
    host_kwargs: ClassVar[dict[str, Any]] = {
        "_scheduler": "nomad",
        "_scheduler_options": {
            "storage_region": "us-east",
        },
        "resolver": "uv",
        "keep_alive": 60,
    }
    app_name: ClassVar[str]
    app_auth: ClassVar[Literal["private", "public", "shared", None]] = None
    request_timeout: ClassVar[int | None] = None
    startup_timeout: ClassVar[int | None] = None

    isolate_channel: async_grpc.Channel | None = None

    def __init_subclass__(cls, **kwargs):
        app_name = kwargs.pop("name", None) or _to_fal_app_name(cls.__name__)
        parent_settings = getattr(cls, "host_kwargs", {})
        cls.host_kwargs = {**parent_settings, **kwargs}

        if cls.request_timeout is not None:
            cls.host_kwargs["request_timeout"] = cls.request_timeout

        if cls.startup_timeout is not None:
            cls.host_kwargs["startup_timeout"] = cls.startup_timeout

        cls.app_name = getattr(cls, "app_name", app_name)

        if cls.__init__ is not App.__init__:
            raise ValueError(
                "App classes should not override __init__ directly. "
                "Use setup() instead."
            )

    def __init__(self, *, _allow_init: bool = False):
        if not _allow_init and not os.getenv("IS_ISOLATE_AGENT"):
            raise NotImplementedError(
                "Running apps through SDK is not implemented yet."
            )

    @classmethod
    def get_endpoints(cls) -> list[str]:
        return [
            signature.path
            for _, endpoint in inspect.getmembers(cls, inspect.isfunction)
            if (signature := getattr(endpoint, "route_signature", None))
        ]

    def collect_routes(self) -> dict[RouteSignature, Callable[..., Any]]:
        return {
            signature: endpoint
            for _, endpoint in inspect.getmembers(self, inspect.ismethod)
            if (signature := getattr(endpoint, "route_signature", None))
        }

    @asynccontextmanager
    async def lifespan(self, app: fastapi.FastAPI):
        _print_python_packages()
        await _call_any_fn(self.setup)
        try:
            yield
        finally:
            await _call_any_fn(self.teardown)

    def health(self):
        return {"version": self.version}

    def setup(self):
        """Setup the application before serving."""

    def teardown(self):
        """Teardown the application after serving."""

    def _add_extra_middlewares(self, app: fastapi.FastAPI):
        @app.middleware("http")
        async def provide_hints_headers(request, call_next):
            response = await call_next(request)
            try:
                # make sure the hints can be encoded in latin-1, so we don't crash
                # when serving.
                # https://github.com/encode/starlette/blob/a766a58d14007f07c0b5782fa78cdc370b892796/starlette/datastructures.py#L568
                hints = []
                for hint in self.provide_hints():
                    try:
                        _ = hint.encode("latin-1")
                        hints.append(hint)
                    except UnicodeEncodeError:
                        from fastapi.logger import logger

                        logger.warning(
                            "Ignoring hint %s for %s because it can't be encoded in "
                            "latin-1",
                            hint,
                            self.__class__.__name__,
                        )

                response.headers["X-Fal-Runner-Hints"] = ",".join(hints)
            except NotImplementedError:
                # This lets us differentiate between apps that don't provide hints
                # and apps that provide empty hints.
                pass
            except Exception:
                from fastapi.logger import logger

                logger.exception(
                    "Failed to provide hints for %s",
                    self.__class__.__name__,
                )
            return response

        @app.middleware("http")
        async def set_global_object_preference(request, call_next):
            try:
                preference_dict = request_lifecycle_preference(request)
                if preference_dict is not None:
                    # This will not work properly for apps with multiplexing enabled
                    # we may mix up the preferences between requests
                    LIFECYCLE_PREFERENCE.set(preference_dict)
            except Exception:
                from fastapi.logger import logger

                logger.exception(
                    "Failed set a global lifecycle preference %s",
                    self.__class__.__name__,
                )

            try:
                return await call_next(request)
            finally:
                # We may miss the global preference if there are operations
                # being done in the background that go beyond the request
                LIFECYCLE_PREFERENCE.set(None)

        @app.middleware("http")
        async def set_request_id(request, call_next):
            # NOTE: Setting request_id is not supported for websocket/realtime endpoints
            if not os.getenv("IS_ISOLATE_AGENT") or not os.environ.get(
                "NOMAD_ALLOC_PORT_grpc"
            ):
                # If not running in the expected environment, skip setting request_id
                return await call_next(request)

            if self.isolate_channel is None:
                grpc_port = os.environ.get("NOMAD_ALLOC_PORT_grpc")
                self.isolate_channel = await open_isolate_channel(
                    f"localhost:{grpc_port}"
                )

            if self.isolate_channel is None:
                return await call_next(request)

            request_id = request.headers.get(REQUEST_ID_KEY)
            if request_id is None:
                return await call_next(request)

            await _set_logger_labels(
                {"fal_request_id": request_id}, channel=self.isolate_channel
            )

            async def _unset_at_end():
                await _set_logger_labels({}, channel=self.isolate_channel)  # type: ignore

            try:
                response: fastapi.responses.Response = await call_next(request)
            except BaseException:
                await _unset_at_end()
                raise
            else:
                # We need to wait for the entire response to be sent before
                # we can set the logger labels back to the default.
                background_tasks = fastapi.BackgroundTasks()
                background_tasks.add_task(_unset_at_end)
                if response.background:
                    # We normally have no background tasks, but we should handle it
                    background_tasks.add_task(response.background)
                response.background = background_tasks

                return response

        @app.exception_handler(RequestCancelledException)
        async def value_error_exception_handler(
            request, exc: RequestCancelledException
        ):
            from fastapi.responses import JSONResponse

            # A 499 status code is not an officially recognized HTTP status code,
            # but it is sometimes used by servers to indicate that a client has closed
            # the connection without receiving a response
            return JSONResponse({"detail": str(exc)}, 499)

    def _add_extra_routes(self, app: fastapi.FastAPI):
        @app.get("/health")
        def health():
            return self.health()

    def provide_hints(self) -> list[str]:
        """Provide hints for routing the application."""
        raise NotImplementedError


def endpoint(
    path: str, *, is_websocket: bool = False
) -> Callable[[EndpointT], EndpointT]:
    """Designate the decorated function as an application endpoint."""

    def marker_fn(callable: EndpointT) -> EndpointT:
        if hasattr(callable, "route_signature"):
            raise ValueError(
                f"Can't set multiple routes for the same function: {callable.__name__}"
            )

        callable.route_signature = RouteSignature(path=path, is_websocket=is_websocket)  # type: ignore
        return callable

    return marker_fn


def _fal_websocket_template(
    func: EndpointT,
    route_signature: RouteSignature,
) -> EndpointT:
    # A template for fal's realtime websocket endpoints to basically
    # be a boilerplate for the user to fill in their inference function
    # and start using it.

    import asyncio
    from collections import deque
    from contextlib import suppress

    import msgpack
    from fastapi import WebSocket, WebSocketDisconnect

    async def mirror_input(queue: deque[Any], websocket: WebSocket) -> None:
        while True:
            try:
                raw_input = await asyncio.wait_for(
                    websocket.receive_bytes(),
                    timeout=route_signature.session_timeout,
                )
            except asyncio.TimeoutError:
                return

            input = msgpack.unpackb(raw_input, raw=False)
            if route_signature.input_modal:
                input = route_signature.input_modal(**input)

            queue.append(input)

    async def mirror_output(
        self,
        queue: deque[Any],
        websocket: WebSocket,
    ) -> None:
        loop = asyncio.get_event_loop()
        max_allowed_buffering = route_signature.buffering or 1
        outgoing_messages: asyncio.Queue[bytes] = asyncio.Queue(
            maxsize=max_allowed_buffering * 2  # x2 for outgoing timings
        )

        async def emit(message):
            if isinstance(message, bytes):
                await websocket.send_bytes(message)
            elif isinstance(message, str):
                await websocket.send_text(message)
            else:
                raise TypeError(f"Can't send message of type {type(message)}")

        async def background_emitter():
            while True:
                output = await outgoing_messages.get()
                await emit(output)
                outgoing_messages.task_done()

        emitter = asyncio.create_task(background_emitter())

        while True:
            if not queue:
                await asyncio.sleep(0.05)
                continue

            input = queue.popleft()
            if input is None or emitter.done():
                if not emitter.done():
                    await outgoing_messages.join()
                    emitter.cancel()

                with suppress(asyncio.CancelledError):
                    await emitter
                return None  # End of input

            batch = [input]
            while queue and len(batch) < route_signature.max_batch_size:
                next_input = queue.popleft()
                if hasattr(input, "can_batch") and not input.can_batch(
                    next_input, len(batch)
                ):
                    queue.appendleft(next_input)
                    break
                batch.append(next_input)

            t0 = loop.time()
            if inspect.iscoroutinefunction(func):
                output = await func(self, *batch)
            else:
                output = await loop.run_in_executor(None, func, self, *batch)  # type: ignore
            total_time = loop.time() - t0
            if not isinstance(output, dict):
                # Handle pydantic output modal
                if hasattr(output, "dict"):
                    output = output.dict()
                else:
                    raise TypeError(
                        "Expected a dict or pydantic model as output, got "
                        f"{type(output)}"
                    )

            messages = [
                msgpack.packb(output, use_bin_type=True),
            ]
            if route_signature.emit_timings:
                # We emit x-fal messages in JSON, no matter what the
                # input/output format is.
                timings = {
                    "type": "x-fal-message",
                    "action": "timings",
                    "timing": total_time,
                }
                messages.append(json.dumps(timings, separators=(",", ":")))

            for message in messages:
                try:
                    outgoing_messages.put_nowait(message)
                except asyncio.QueueFull:
                    await emit(message)

    async def websocket_template(self, websocket: WebSocket) -> None:
        import asyncio

        await websocket.accept()

        queue: deque[Any] = deque(maxlen=route_signature.buffering)
        input_task = asyncio.create_task(mirror_input(queue, websocket))
        input_task.add_done_callback(lambda _: queue.append(None))
        output_task = asyncio.create_task(mirror_output(self, queue, websocket))

        try:
            await asyncio.wait(
                {
                    input_task,
                    output_task,
                },
                return_when=asyncio.FIRST_COMPLETED,
            )
            if input_task.done():
                # User didn't send any input within the timeout
                # so we can just close the connection after the
                # processing of the last input is done.
                input_task.result()
                await asyncio.wait_for(
                    output_task, timeout=route_signature.session_timeout
                )
            else:
                assert output_task.done()

                # The execution of the inference function failed or exitted,
                # so just propagate the result.
                input_task.cancel()
                with suppress(asyncio.CancelledError):
                    await input_task

                output_task.result()
        except WebSocketDisconnect:
            input_task.cancel()
            output_task.cancel()
            with suppress(asyncio.CancelledError):
                await input_task

            with suppress(asyncio.CancelledError):
                await output_task
        except Exception as exc:
            import traceback

            traceback.print_exc()

            await websocket.send_json(
                {
                    "type": "x-fal-error",
                    "error": "INTERNAL_ERROR",
                    "reason": str(exc),
                }
            )
        else:
            await websocket.send_json(
                {
                    "type": "x-fal-error",
                    "error": "TIMEOUT",
                    "reason": "no inputs, reconnect when needed!",
                }
            )

        await websocket.close()

    # Seems like templating + stringified annotations don't play well,
    # so we have to set them manually.
    websocket_template.__annotations__ = {
        "websocket": WebSocket,
        "return": None,
    }
    websocket_template.route_signature = route_signature  # type: ignore
    websocket_template.original_func = func  # type: ignore
    return typing.cast(EndpointT, websocket_template)


_SENTINEL = object()


def realtime(
    path: str,
    *,
    buffering: int | None = None,
    session_timeout: float | None = None,
    input_modal: Any | None = _SENTINEL,
    max_batch_size: int = 1,
) -> Callable[[EndpointT], EndpointT]:
    """Designate the decorated function as a realtime application endpoint."""

    def marker_fn(original_func: EndpointT) -> EndpointT:
        nonlocal input_modal

        if hasattr(original_func, "route_signature"):
            raise ValueError(
                "Can't set multiple routes for the same function: "
                f"{original_func.__name__}"
            )

        if input_modal is _SENTINEL:
            type_hints = typing.get_type_hints(original_func)
            if len(type_hints) >= 1:
                input_modal = type_hints[list(type_hints.keys())[0]]
            else:
                input_modal = None

        route_signature = RouteSignature(
            path=path,
            is_websocket=True,
            input_modal=input_modal,
            buffering=buffering,
            session_timeout=session_timeout,
            max_batch_size=max_batch_size,
        )
        return _fal_websocket_template(
            original_func,
            route_signature,
        )

    return marker_fn



from __future__ import annotations

import subprocess
from contextlib import suppress

import pytest
from isolate.backends.common import active_python
from pydantic import __version__ as pydantic_version

import fal
from fal.api import FalServerlessError, Options
from fal.container import ContainerImage
from fal.toolkit.file import File

PACKAGE_NAME = "fall"


def git_revision_short_hash() -> str:
    return (
        subprocess.check_output(["git", "rev-parse", "--short", "HEAD"])
        .decode("ascii")
        .strip()
    )


@pytest.mark.xfail(reason="Temporary mismatch due to grpc version updates. Ping @efiop")
def test_missing_dependencies_nested_server_error(isolated_client):
    @isolated_client()
    def test1():
        return "hello"

    @isolated_client()
    def test2():
        test1()

    with pytest.raises(
        FalServerlessError,
        match=r"accessed through 'test1'",
    ):
        test2()


def test_regular_function(isolated_client):
    @isolated_client("virtualenv")
    def regular_function():
        return 42

    assert regular_function() == 42

    @isolated_client("virtualenv")
    def mult(a, b):
        return a * b

    assert mult(5, 2) == 10


def test_regular_function_on_nomad(isolated_client):
    @isolated_client(_scheduler="nomad")
    def regular_function():
        return 42

    assert regular_function() == 42

    @isolated_client(_scheduler="nomad")
    def mult(a, b):
        return a * b

    assert mult(5, 2) == 10


def test_regular_function_in_a_container(isolated_client):
    @isolated_client("container")
    def regular_function():
        return 42

    assert regular_function() == 42

    @isolated_client("container")
    def mult(a, b):
        return a * b

    assert mult(5, 2) == 10


def test_regular_function_in_a_container_with_custom_image(isolated_client):
    actual_python = active_python()

    @isolated_client(
        "container",
        image=ContainerImage.from_dockerfile_str(
            f"""
            FROM python:{actual_python}-slim
            # {git_revision_short_hash()}
            RUN env
            """
        ),
    )
    def regular_function():
        return 42

    assert regular_function() == 42

    @isolated_client(
        "container",
        image=ContainerImage.from_dockerfile_str(
            f"FROM python:{actual_python}-slim\n# {git_revision_short_hash()}"
        ),
    )
    def mult(a, b):
        return a * b

    assert mult(5, 2) == 10


def test_function_pipelining(isolated_client):
    @isolated_client("virtualenv")
    def regular_function():
        return 42

    @isolated_client("virtualenv")
    def calling_function(result):
        return result * 2

    assert calling_function(regular_function()) == 84


def test_function_exception(isolated_client):
    class FooException(Exception):
        pass

    class BarException(Exception):
        pass

    class MyFuncException(Exception):
        pass

    def foo():
        raise FooException("foo")

    def bar():
        try:
            foo()
        except Exception as exc:
            raise BarException("bar") from exc

    @isolated_client()
    def myfunc():
        try:
            bar()
        except Exception as exc:
            raise MyFuncException("myfunc") from exc

    with pytest.raises(MyFuncException) as excinfo:
        myfunc()

    bar_exc = excinfo.value.__cause__
    assert isinstance(bar_exc, BarException)
    assert isinstance(bar_exc.__cause__, FooException)


@pytest.mark.xfail(reason="See https://github.com/fal-ai/fal/issues/169")
def test_function_calling_other_function(isolated_client):
    try:
        import importlib.metadata as importlib_metadata
    except ImportError:
        import importlib_metadata

    package_version = importlib_metadata.version(PACKAGE_NAME)

    @isolated_client("virtualenv")
    def regular_function():
        return 42

    @isolated_client(
        "virtualenv",
        requirements=[
            f"{PACKAGE_NAME}>=0.6.19,<={package_version}",
        ],
    )
    def calling_function(recurse):
        import os

        for name in os.environ:
            if name.startswith("FAL_"):
                print(os.environ[name])

        if recurse:
            return calling_function(recurse=False)
        else:
            return regular_function()

    # One level of direct calls
    assert calling_function(recurse=False) == 42

    # Two levels of direct calls
    assert calling_function(recurse=True) == 42


@pytest.mark.xfail
def test_dependency_inference(isolated_client):
    @isolated_client("virtualenv")
    def regular_function():
        return 42

    @isolated_client("virtualenv")
    def calling_function():
        return regular_function()

    assert calling_function() == 42


@pytest.mark.xfail
def test_process_crash(isolated_client):
    # We can catch the SystemExit if it originates from
    # an exception (in the case of exit(1)).
    @isolated_client("virtualenv")
    def process_crash_regular():
        exit(1)

    with pytest.raises(SystemExit):
        process_crash_regular()

    # But if it is a native exit call, then the agent process
    # will crash and there is not much we can do other than offer
    # a decent error message.
    @isolated_client("virtualenv")
    def process_crash_without_catching():
        import os

        os._exit(0)

    with pytest.raises(FalServerlessError, match="..."):
        process_crash_without_catching()


@pytest.mark.xfail
def test_unserializable_input_function(isolated_client):
    # When the function can not be serialized at all (has a reference to an unpicklable
    # object (like a frame)).

    @isolated_client("virtualenv")
    def unpicklable_input_function_client(default=__import__("sys")._getframe(0)):
        return default

    with pytest.raises(FalServerlessError, match="..."):
        unpicklable_input_function_client()


@pytest.mark.xfail
def test_unserializable_return(isolated_client):
    # When the return of the function can not be serialized.

    @isolated_client("virtualenv")
    def unpicklable_return():
        import sys

        return sys._getframe(0)

    with pytest.raises(FalServerlessError, match="..."):
        unpicklable_return()


@pytest.mark.xfail
def test_missing_dependencies_on_client(isolated_client):
    # When the return can't be deserialized on the client side due to
    # incomplete/missing dependencies.

    @isolated_client("virtualenv", requirements=["refactor"])
    def unpicklable_input_function_client():
        import refactor

        return refactor.BaseAction()

    with pytest.raises(FalServerlessError, match="..."):
        unpicklable_input_function_client()


@pytest.mark.xfail
def test_when_exception_can_not_be_deserialized(isolated_client):
    # When the exception can't be deserialized on the client side due to
    # incomplete/missing dependencies.

    @isolated_client("virtualenv", requirements=[""])
    def unpicklable_input_function_client():
        class T(Exception):
            frame = __import__("sys")._getframe(0)

        raise T()

    with pytest.raises(FalServerlessError, match="..."):
        unpicklable_input_function_client()


@pytest.mark.xfail
def test_client_superseeding_dependencies_crash(isolated_client):
    # E.g. I depend on tensorboard which depends on dill==0.3.2
    # but the agent overrides it with dill==0.3.5.1 so I get an
    # unexpected error at runtime (instead of at environment creation
    # time)

    @isolated_client("virtualenv", requirements=["dill==0.3.2"])
    def conflicting_environment():
        import dill

        assert dill.__version__ == "0.3.2"

    # This should crash at the build stage
    with pytest.raises(
        FalServerlessError, match="package versions have conflicting dependencies."
    ):
        conflicting_environment()


@pytest.mark.xfail
def test_memory_overflow_crash_on_run(isolated_client):
    @isolated_client("virtualenv")
    def memory_overflow_crash_on_run():
        import os

        objects = []
        while True:
            objects.append(os.urandom(1024**3))

    with pytest.raises(Exception, match="Insufficient memory"):
        memory_overflow_crash_on_run()


def test_keepalive_after_agent_exit(isolated_client):
    # Should work (fresh)
    @isolated_client("virtualenv")
    def regular_function():
        return 42

    assert regular_function() == 42

    # This uses the cached agent process but it will
    # kill it at the end.
    with suppress(BaseException):

        @isolated_client("virtualenv")
        def process_crash_without_catching():
            import os

            os._exit(0)

        process_crash_without_catching()

    # This should start a fresh agent process.
    @isolated_client("virtualenv")
    def regular_function():
        return 42

    assert regular_function() == 42


def test_faulty_setup_function(isolated_client):
    def good_setup_function():
        return 42

    def bad_setup_function():
        raise ValueError()

    @isolated_client("virtualenv")
    def regular_function(result):
        return result * 2

    first_case = regular_function.on(setup_function=good_setup_function)
    assert first_case() == 84

    second_case = regular_function.on(setup_function=bad_setup_function)
    with pytest.raises(ValueError):
        second_case()


def test_unserializable_setup_function(isolated_client):
    def unpicklable_setup_function():
        import sys

        return sys._getframe(0)

    @isolated_client("virtualenv", setup_function=unpicklable_setup_function)
    def regular_function(result):
        return str(result)

    # This works fine since the setup function's result is not returned
    # to the client (but rather just kept on the server).
    assert isinstance(regular_function(), str)


def test_big_message(isolated_client):
    # Default gRPC max message size is 4MB so let's
    # try doubling that.
    data_length = 8 * (1024**2)

    @isolated_client("virtualenv", machine_type="M")
    def big_input_function(data):
        return len(data)

    # Small inputs should work fine (pre-test).
    assert big_input_function(b"0") == 1
    assert big_input_function(b"0" * data_length) == data_length

    # Try receiving a big message.
    @isolated_client("virtualenv", machine_type="M")
    def big_return_function(data_length):
        return b"0" * data_length

    assert len(big_return_function(1)) == 1
    assert len(big_return_function(data_length)) == data_length


def test_futures(isolated_client):
    from concurrent.futures import wait

    @isolated_client("virtualenv")
    def regular_function(n):
        return n * 2

    future = regular_function.submit(1)
    assert future.result() == 2

    future_2 = regular_function.submit(2)
    future_3 = regular_function.submit(3)
    future_4 = regular_function.submit(4)
    wait([future_2, future_3, future_4])

    # Ensure the futures are done in the correct order
    # (no mix and match between arguments and the returned
    # future).
    assert future_2.result() == 4
    assert future_3.result() == 6
    assert future_4.result() == 8


@pytest.mark.xfail(reason="Nomad does not support conda")
def test_conda_environment_on_nomad(isolated_client):
    @isolated_client(
        "conda",
        packages=["pyjokes=0.6.0"],
        machine_type="L",
        # conda is only supported on Kubernetes
        _scheduler="nomad",
    )
    def regular_function():
        import pyjokes

        return pyjokes.__version__

    assert regular_function() == "0.6.0"


@pytest.mark.xfail(reason="Broken on nomad")
def test_cached_function(isolated_client, capsys, monkeypatch):
    import inspect
    import time

    test_stamp = time.time()
    real_getsource = inspect.getsource

    # For ensuring between test runs the cached functions
    # are gone, we add a timestamp to the source code.
    def add_timestamp_to_source(func):
        return real_getsource(func) + f"\n# {test_stamp}"

    monkeypatch.setattr(inspect, "getsource", add_timestamp_to_source)

    @fal.cached
    def get_pipe():
        import time

        print("computing")
        time.sleep(1)  # slow IO
        return "pipe"

    @fal.cached
    def factorial(n: int) -> int:
        import math
        import time

        print("computing")
        time.sleep(1)  # slow CPU
        return math.factorial(n)

    # make sure this machine is not shared with others by using a unique requirements
    @isolated_client(
        "virtualenv",
        keep_alive=30,
        requirements=["pyjokes     "],
    )
    def regular_function(n):
        if get_pipe() == "pipe":
            return factorial(n)

    assert regular_function(4) == 24
    out, err = capsys.readouterr()
    assert out.count("computing") == 2

    # pipe is now cached, but using a different factorial(n)
    assert regular_function(3) == 6
    out, err = capsys.readouterr()
    assert out.count("computing") == 1

    # This should be all cached
    assert regular_function(4) == 24
    out, err = capsys.readouterr()
    assert out.count("computing") == 0

    # So does thiss
    assert regular_function(3) == 6
    out, err = capsys.readouterr()
    assert out.count("computing") == 0

    # But this should not be cached
    assert regular_function(5) == 120
    out, err = capsys.readouterr()
    assert out.count("computing") == 1


def test_pydantic_serialization(isolated_client):
    from pydantic import BaseModel, Field

    class MathQuery(BaseModel):
        x: int = Field(gt=0, description="The first operand")
        y: int = Field(gt=0, description="The second operand")

    class MathResult(BaseModel):
        result: int = Field(description="The result of the operation")

    @isolated_client("virtualenv", requirements=[f"pydantic=={pydantic_version}"])
    def add(query: MathQuery) -> MathResult:
        return MathResult(result=query.x + query.y)

    result = add(MathQuery(x=1, y=2))
    assert result.result == 3
    assert result == MathResult(result=3)


def test_serve_on_off(isolated_client):
    from pydantic import BaseModel, Field

    class MathQuery(BaseModel):
        x: int = Field(gt=0, description="The first operand")
        y: int = Field(gt=0, description="The second operand")

    class MathResult(BaseModel):
        result: int = Field(description="The result of the operation")

    @isolated_client(
        "virtualenv", serve=True, requirements=[f"pydantic=={pydantic_version}"]
    )
    def add(query: MathQuery) -> MathResult:
        return MathResult(result=query.x + query.y)

    local_first_add = add.on(serve=False)
    assert local_first_add(MathQuery(x=1, y=2)) == MathResult(result=3)

    local_first_add_2 = local_first_add.on(serve=True).on(serve=False)
    assert local_first_add_2(MathQuery(x=1, y=2)) == MathResult(result=3)


def test_worker_env_vars(isolated_client):
    from fal.flags import GRPC_HOST

    @isolated_client("virtualenv", keep_alive=5)
    def get_env_var(name: str) -> str | None:
        import os

        return os.getenv(name, None)

    fal_host = get_env_var("FAL_HOST")
    assert fal_host, "FAL_HOST is not set"
    assert fal_host == GRPC_HOST, "FAL_HOST is not set to the correct value"
    print(f"FAL_HOST: {fal_host}, GRPC_HOST: {GRPC_HOST}")

    fal_key_id = get_env_var("FAL_KEY_ID")
    assert fal_key_id, "FAL_KEY_ID is not set"

    fal_key_secret = get_env_var("FAL_KEY_SECRET")
    assert fal_key_secret, "FAL_KEY_SECRET is not set"


@pytest.mark.parametrize(
    "repo_type, url_prefix",
    [
        (
            "fal",
            "https://storage.googleapis.com/isolate-dev-smiling-shark_toolkit_bucket/",
        ),
        ("fal_v2", "https://v2.fal.media/files"),
        ("fal_v3", "https://v3.fal.media/files"),
    ],
)
def test_fal_storage(isolated_client, repo_type, url_prefix):
    file = File.from_bytes(
        b"Hello fal storage from local", repository=repo_type, fallback_repository=None
    )
    assert file.url.startswith(url_prefix)
    assert file.as_bytes().decode().endswith("local")

    @isolated_client(
        serve=True, requirements=[f"pydantic=={pydantic_version}", "tomli"]
    )
    def hello_file():
        # Run in the isolated environment
        return File.from_bytes(
            b"Hello fal storage from isolated",
            repository=repo_type,
            save_kwargs={
                "multipart": False,
                "multipart_threshold": 1024 * 1024,
                "multipart_chunk_size": 1024 * 1024,
                "multipart_max_concurrency": 10,
                "object_lifecycle_preference": {},
            },
            fallback_repository=None,
        )

    local_fn = hello_file.on(serve=False)
    file = local_fn()
    assert file.url.startswith(url_prefix)
    assert file.as_bytes().decode().endswith("isolated")


def test_on_changes():
    def fastapi_in(op: Options):
        return any("fastapi" in req for req in op.environment["requirements"])

    @fal.function(
        requirements=["pyjokes"],
        serve=True,
        keep_alive=10,
        machine_type="M",
        max_concurrency=8,
    )
    def served(name: str):
        return File.from_bytes(f"Hello {name} from isolated".encode(), repository="fal")

    op = served.options
    assert fastapi_in(op), "fastapi not in served environment"
    assert op.gateway["serve"], "serve not in served gateway"
    assert op.host["max_concurrency"] == 8, "max_concurrency not in served gateway"
    assert op.host["keep_alive"] == 10, "keep_alive not in served host"
    assert op.host["machine_type"] == "M", "machine_type not in served host"

    unserved = served.on(serve=False)
    # should only change serve value
    op = unserved.options
    assert fastapi_in(op), "fastapi MUST stay after removing serve"
    assert not op.gateway["serve"], "serve STILL set in unserved gateway"
    assert op.host["max_concurrency"] == 8, "max_concurrency not in unserved gateway"
    assert op.host["keep_alive"] == 10, "keep_alive not in unserved host"
    assert op.host["machine_type"] == "M", "machine_type not in unserved host"

    local = served.on(host=fal.local)
    # should remove all host values
    op = local.options
    assert fastapi_in(op), "fastapi not in local environment"
    assert op.gateway["serve"], "serve not in local gateway"
    assert "max_concurrency" not in op.host, "max_concurrency set in local gateway"
    assert "keep_alive" not in op.host, "keep_alive set in local host"
    assert "machine_type" not in op.host, "machine_type set in local host"
